/*
 * Example 46-2
 * The Clock widget data structure.
 */

#include "tk.h"
#include <sys/time.h>

typedef struct {
	Tk_Window tkwin;							/* The window for the widget */
	Display *display;							/* Tk's handle on the display */
	Tcl_Interp *interp;							/* Interpreter of the widget */
	Tcl_Command widgetCmd;							/* clock instance command. */
	Tk_OptionTable optionTable;		/* Used to parse configuration options */
	/*
	 * Clock-specific attributes.
	 */
	int borderWidth;							/* Size of 3-D border */
	Tcl_Obj *borderWidthPtr;	/* Original string value */
	int relief;							/* Style of 3-D border */
	Tk_3DBorder background;							/* Color for border & background */
	XColor *foreground;							/* Color for the text */
	XColor *highlight;							/* Color for active highlight */
	XColor *highlightBg;							/* Color for neutral highlight */
	int highlightWidth;							/* Thickness of highlight rim */
	Tcl_Obj *highlightWidthPtr; /* Original string value */
	Tk_Font tkfont;							/* Font info for the text */
	char *format;							/* Format for time string */
	/*
	 * Graphic contexts and other support.
	 */
	GC textGC;							/* Text graphics context */
	Tk_TimerToken token;							/* Periodic callback handle*/
	char *clock;							/* Pointer to the clock string */
	int numChars;							/* length of the text */
	int textWidth;							/* in pixels */
	Tcl_Obj *widthPtr;							/* The original width string value*/
	int textHeight;							/* in pixels */
	Tcl_Obj *heightPtr;							/* The original height string value*/
	int padX;							/* Horizontal padding */
	Tcl_Obj *padXPtr;							/* The original padX string value*/
	int padY;							/* Vertical padding */
	Tcl_Obj *padYPtr;							/* The original padY string value */
	int flags;							/* Flags defined below */
} Clock;
/*
* Flag bit definitions.
*/
#define REDRAW_PENDING							0x1
#define GOT_FOCUS							0x2
#define TICKING							0x4


/*
 * Example 46-10
 * Configuration specs for the clock widget.
 */

static Tk_ConfigSpec configSpecs[] = {
	{TK_CONFIG_BORDER, "-background", "background",
		"Background", "light blue",
		Tk_Offset(Clock, background), TK_CONFIG_COLOR_ONLY},
	{TK_CONFIG_BORDER, "-background", "background",
		"Background", "white", Tk_Offset(Clock, background),
		TK_CONFIG_MONO_ONLY},
	{TK_CONFIG_SYNONYM, "-bg", "background", (char *) NULL,
		(char *) NULL, 0, 0},
	{TK_CONFIG_SYNONYM, "-bd", "borderWidth", (char *) NULL,
		(char *) NULL, 0, 0},
	{TK_CONFIG_PIXELS, "-borderwidth", "borderWidth",
		"BorderWidth","2", Tk_Offset(Clock, borderWidth), 0},
	{TK_CONFIG_RELIEF, "-relief", "relief", "Relief",
		"ridge", Tk_Offset(Clock, relief), 0},
	{TK_CONFIG_COLOR, "-foreground", "foreground",
		"Foreground", "black", Tk_Offset(Clock, foreground),
		0},
	{TK_CONFIG_SYNONYM, "-fg", "foreground", (char *) NULL,
		(char *) NULL, 0, 0},
	{TK_CONFIG_COLOR, "-highlightcolor", "highlightColor",
		"HighlightColor", "red", Tk_Offset(Clock, highlight),
		TK_CONFIG_COLOR_ONLY},
	{TK_CONFIG_COLOR, "-highlightcolor", "highlightColor",
		 "HighlightColor", "black", 
		Tk_Offset(Clock, highlight),TK_CONFIG_MONO_ONLY},
	{TK_CONFIG_COLOR, "-highlightbackground", 
		"highlightBackground", "HighlightBackground",
		"light blue", Tk_Offset(Clock, highlightBg),
		TK_CONFIG_COLOR_ONLY},
	{TK_CONFIG_COLOR, "-highlightbackground", 
		"highlightBackground", "HighlightBackground",
		"black", Tk_Offset(Clock, highlightBg),
		TK_CONFIG_MONO_ONLY},
	{TK_CONFIG_PIXELS, "-highlightthickness",
		"highlightThickness","HighlightThickness",
		"2", Tk_Offset(Clock, highlightWidth), 0},
	{TK_CONFIG_PIXELS, "-padx", "padX", "Pad",
		"2", Tk_Offset(Clock, padX), 0},
	{TK_CONFIG_PIXELS, "-pady", "padY", "Pad",
		"2", Tk_Offset(Clock, padY), 0},
	{TK_CONFIG_STRING, "-format", "format", "Format",
		"%H:%M:%S", Tk_Offset(Clock, format), 0},
	{TK_CONFIG_FONT, "-font", "font", "Font",
		"Courier 18",
		 Tk_Offset(Clock, tkfont), 0},
	{TK_CONFIG_END, (char *) NULL, (char *) NULL,
		 (char *) NULL, (char *) NULL, 0, 0}
};


/*
 * Example 46-13
 * ComputeGeometry computes the widget\" s size.
 */

static void
ComputeGeometry(Clock *clockPtr)
{
	int width, height;
	Tk_FontMetrics fm;							/* Font size information */
	struct tm *tmPtr;							/* Time info split into fields */
	struct timeval tv;							/* BSD-style time value */
	int bd;							/* Padding from borders */
	char clock[1000];							/* Displayed time */

	/*
	 * Get the time and format it to see how big it will be.
	 */
	gettimeofday(&tv, NULL);
	tmPtr = localtime(&tv.tv_sec);
	strftime(clock, 1000, clockPtr->format, tmPtr);
	if (clockPtr->clock != NULL) {
	    Tcl_Free(clockPtr->clock);
	}
	clockPtr->clock = ckalloc(1+strlen(clock));
	clockPtr->numChars = strlen(clock);

	bd = clockPtr->highlightWidth + clockPtr->borderWidth;
	Tk_GetFontMetrics(clockPtr->tkfont, &fm);
	height = fm.linespace + 2*(bd + clockPtr->padY);
	Tk_MeasureChars(clockPtr->tkfont, clock,
		clockPtr->numChars, 0, 0, &clockPtr->textWidth);
	width = clockPtr->textWidth + 2*(bd + clockPtr->padX);

	Tk_GeometryRequest(clockPtr->tkwin, width, height);
	Tk_SetInternalBorder(clockPtr->tkwin, bd);
}


/*
 * Example 46-14
 * The ClockDisplay procedure.
 */

static void
ClockDisplay(ClientData clientData)
{
	Clock *clockPtr = (Clock *)clientData;
	Tk_Window tkwin = clockPtr->tkwin;
	GC gc;							/* Graphics Context for highlight */
	Tk_TextLayout layout;							/* Text measurement state */
	Pixmap pixmap;							/* Temporary drawing area */
	int offset, x, y;							/* Coordinates */
	int width, height;							/* Size */
	struct tm *tmPtr;							/* Time info split into fields */
	struct timeval tv;							/* BSD-style time value */

	/*
	 * Make sure the clock still exists
	 * and is mapped onto the display before painting.
	 */
	clockPtr->flags &= ~(REDRAW_PENDING|TICKING);
	if ((clockPtr->tkwin == NULL) || !Tk_IsMapped(tkwin)) {
		return;
	}
	/*
	 * Format the time into a string.
	 * localtime chops up the time into fields.
	 * strftime formats the fields into a string.
	 */
	gettimeofday(&tv, NULL);
	tmPtr = localtime(&tv.tv_sec);
	strftime(clockPtr->clock, clockPtr->numChars+1, 
		clockPtr->format, tmPtr);
	/*
	 * To avoid flicker when the display is updated, the new
	 * image is painted in an offscreen pixmap and then 
	 * copied onto the display in one operation. Allocate the
	 * pixmap and paint its background.
	 */
	pixmap = Tk_GetPixmap(clockPtr->display,
		Tk_WindowId(tkwin), Tk_Width(tkwin),
		Tk_Height(tkwin), Tk_Depth(tkwin));
	Tk_Fill3DRectangle(tkwin, pixmap, 
		clockPtr->background, 0, 0, Tk_Width(tkwin),
		Tk_Height(tkwin), 0, TK_RELIEF_FLAT);

	/*
	 * Paint the text first.
	 */
	layout = Tk_ComputeTextLayout(clockPtr->tkfont, 
		clockPtr->clock, clockPtr->numChars, 0,
		TK_JUSTIFY_CENTER, 0, &width, &height);
	x = (Tk_Width(tkwin) - width)/2;
	y = (Tk_Height(tkwin) - height)/2;
	Tk_DrawTextLayout(clockPtr->display, pixmap,
		clockPtr->textGC, layout, x, y, 0, -1);

	/*
	 * Display the borders, so they overwrite any of the
	 * text that extends to the edge of the display.
	 */
	if (clockPtr->relief != TK_RELIEF_FLAT) {
		Tk_Draw3DRectangle(tkwin, pixmap,
			clockPtr->background,
			clockPtr->highlightWidth,
			clockPtr->highlightWidth,
			Tk_Width(tkwin) - 2*clockPtr->highlightWidth,
			Tk_Height(tkwin) - 2*clockPtr->highlightWidth,
			clockPtr->borderWidth, clockPtr->relief);
	}
	if (clockPtr->highlightWidth != 0) {
		GC gc;

		/*
		 * This GC is associated with the color, and Tk caches
		 * the GC until the color is freed. Hence no freeGC.
		 */

		if (clockPtr->flags & GOT_FOCUS) {
			gc = Tk_GCForColor(clockPtr->highlight, pixmap);
		} else {
			gc = Tk_GCForColor(clockPtr->highlightBg, pixmap);
		}
		Tk_DrawFocusHighlight(tkwin, gc, 
			clockPtr->highlightWidth, pixmap);
	}
	/*
	 * Copy the information from the off-screen pixmap onto
	 * the screen, then delete the pixmap.
	 */
    
	XCopyArea(clockPtr->display, pixmap, Tk_WindowId(tkwin),
		clockPtr->textGC, 0, 0, Tk_Width(tkwin),
		Tk_Height(tkwin), 0, 0);
	Tk_FreePixmap(clockPtr->display, pixmap);

	/*
	 * Queue another call to ourselves. The rate at which
	 * this is done could be optimized.
	 */
	clockPtr->token = Tk_CreateTimerHandler(1000, 
		ClockDisplay, (ClientData)clockPtr);
	clockPtr->flags |= TICKING;
}


/*
 * Example 46-16
 * The ClockDestroy cleanup procedure.
 */

static void
ClockDestroy(clientData)
	ClientData clientData;		/* Info about entry widget. */
{
	register Clock *clockPtr = (Clock *) clientData;
 
	/*
	 * Free up all the stuff that requires special handling,
	 * then let Tk_FreeOptions handle resources associated
	 * with the widget attributes.
	 */
	if (clockPtr->textGC != None) {
		Tk_FreeGC(clockPtr->display, clockPtr->textGC);
	}
	if (clockPtr->clock != NULL) {
		Tcl_Free(clockPtr->clock);
	}
	if (clockPtr->flags & TICKING) {
		Tk_DeleteTimerHandler(clockPtr->token);
	}
	if (clockPtr->flags & REDRAW_PENDING) {
		Tk_CancelIdleCall(ClockDisplay,
			(ClientData) clockPtr);
	}
	/*
	 * This frees up colors and fonts and any allocated
	 * storage associated with the widget attributes.
	 */
	Tk_FreeOptions(configSpecs, (char *) clockPtr,
		clockPtr->display, 0);
	Tcl_Free((char *) clockPtr);
}


/*
 * Example 46-15
 * The ClockEventProc handles window events.
 */

static void
ClockEventProc(ClientData clientData, XEvent *eventPtr)
{
	Clock *clockPtr = (Clock *) clientData;
	if ((eventPtr->type == Expose) &&
		(eventPtr->xexpose.count == 0)) {
			goto redraw;
	} else if (eventPtr->type == DestroyNotify) {
		Tcl_DeleteCommandFromToken(clockPtr->interp, 
				clockPtr->widgetCmd);
		/*
		 * Zapping the tkwin lets the other procedures
		 * know we are being destroyed.
		 */
		clockPtr->tkwin = NULL;

		if (clockPtr->flags & REDRAW_PENDING) {
			Tk_CancelIdleCall(ClockDisplay,
				(ClientData) clockPtr);
			clockPtr->flags &= ~REDRAW_PENDING;
		}
		if (clockPtr->flags & TICKING) {
			Tk_DeleteTimerHandler(clockPtr->token);
			clockPtr->flags &= ~TICKING;
		}
		/*
		 * This results in a call to ClockDestroy.
		 */
		Tk_EventuallyFree((ClientData) clockPtr,
			ClockDestroy);
	} else if (eventPtr->type == FocusIn) {
		if (eventPtr->xfocus.detail != NotifyPointer) {
			clockPtr->flags |= GOT_FOCUS;
			if (clockPtr->highlightWidth > 0) {
				goto redraw;
			}
		}
	} else if (eventPtr->type == FocusOut) {
		if (eventPtr->xfocus.detail != NotifyPointer) {
			clockPtr->flags &= ~GOT_FOCUS;
			if (clockPtr->highlightWidth > 0) {
				goto redraw;
			}
		}
	}
	return;
redraw:
	if ((clockPtr->tkwin != NULL) &&
			!(clockPtr->flags & REDRAW_PENDING)) {
		Tk_DoWhenIdle(ClockDisplay, (ClientData) clockPtr);
		clockPtr->flags |= REDRAW_PENDING;
	}
}


/*
 * Example 46-7
 * ClockConfigure allocates resources for the widget.
 */

static int
ClockConfigure(interp, clockPtr, argc, argv, flags)
	Tcl_Interp *interp;	/* For return values and errors */
	Clock *clockPtr;					/* The per-instance data structure */
	int argc;					/* Number of valid entries in argv */
	char *argv[];					/* The command line arguments */
	int flags;					/* Tk_ConfigureWidget flags */
{
	XGCValues gcValues;
	GC newGC;

	/*
	 * Tk_ConfigureWidget parses the command line arguments
	 * and looks for defaults in the resource database.
	 */
	if (Tk_ConfigureWidget(interp, clockPtr->tkwin,
			configSpecs, argc, argv, (char *) clockPtr, flags)
				!= TCL_OK) {
		return TCL_ERROR;
	}
	/*
	 * Give the widget a default background so it doesn\" t get
	 * a random background between the time it is initially
	 * displayed by the X server and we paint it
	 */
	Tk_SetWindowBackground(clockPtr->tkwin,
		Tk_3DBorderColor(clockPtr->background)->pixel);
	/*
	 * Set up the graphics contexts to display the widget.
	 * The context is used to draw off-screen pixmaps,
	 * so turn off exposure notifications.
	 */
	gcValues.background = 
		Tk_3DBorderColor(clockPtr->background)->pixel;
	gcValues.foreground = clockPtr->foreground->pixel;
	gcValues.font = Tk_FontId(clockPtr->tkfont);
	gcValues.graphics_exposures = False;
	newGC = Tk_GetGC(clockPtr->tkwin,
		GCBackground|GCForeground|GCFont|GCGraphicsExposures,
		&gcValues);
	if (clockPtr->textGC != None) {
	    Tk_FreeGC(clockPtr->display, clockPtr->textGC);
	}
	clockPtr->textGC = newGC;
	/*
	 * Determine how big the widget wants to be.
	 */
	ComputeGeometry(clockPtr);
	/*
	 * Set up a call to display ourself.
	 */
	if ((clockPtr->tkwin != NULL) &&
		 	 Tk_IsMapped(clockPtr->tkwin)
		 	&& !(clockPtr->flags & REDRAW_PENDING)) {
		Tk_DoWhenIdle(ClockDisplay, (ClientData) clockPtr);
		clockPtr->flags |= REDRAW_PENDING;
	}
	return TCL_OK;
}


/*
 * Example 46-5
 * The ClockInstanceCmd command procedure.
 */

static int
ClockInstanceCmd(clientData, interp, argc, argv)
	ClientData clientData;						/* A pointer to a Clock struct */
	Tcl_Interp *interp;						/* The interpreter */
	int argc;						/* The number of arguments */
	char *argv[];						/* The command line arguments */
{
	Clock *clockPtr = (Clock *)clientData;
	int result = TCL_OK;
	char c;
	int len;

	if (argc < 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], " option ?arg arg ...?\"", (char *) NULL);
		return TCL_ERROR;
	}
	c = argv[1][0];
	len = strlen(argv[1]);
	if ((c == 'c') && (strncmp(argv[1], "cget", len) == 0)
			&& (len >= 2)) {
		if (argc != 3) {
			Tcl_AppendResult(interp,
				 "wrong # args: should be \"",
				argv[0], " cget option\"",
				(char *) NULL);
			return TCL_ERROR;
		}
		result = Tk_ConfigureValue(interp, clockPtr->tkwin,
			 configSpecs, (char *) clockPtr, argv[2], 0);
	} else if ((c == 'c') && (strncmp(argv[1], "configure", len) 
				== 0) && (len >= 2)) {
		if (argc == 2) {
			/*
			 * Return all configuration information.
			 */
			result = Tk_ConfigureInfo(interp, clockPtr->tkwin,
				configSpecs, (char *) clockPtr,
				(char *) NULL,0);
		} else if (argc == 3) {
			/*
			 * Return info about one attribute, like cget.
			 */
			result = Tk_ConfigureInfo(interp, clockPtr->tkwin,
				configSpecs, (char *) clockPtr, argv[2], 0);
		} else {
			/*
			 * Change one or more attributes.
			 */
			result = ClockConfigure(interp, clockPtr, argc-2,
				argv+2,TK_CONFIG_ARGV_ONLY);
		}
	} else {
		Tcl_AppendResult(interp, "bad option \"", argv[1],
			"\": must be cget, configure, position, or size",
			(char *) NULL);
		return TCL_ERROR;
	}
	return result;
}


/*
 * Example 46-3
 * The ClockCmd command procedure.
 */

int
ClockCmd(clientData, interp, argc, argv)
	ClientData clientData;						/* Main window of the app */
	Tcl_Interp *interp;						/* Current interpreter. */
	int argc;						/* Number of arguments. */
	char **argv;						/* Argument strings. */
{
	Tk_Window main = (Tk_Window) clientData;
	Clock *clockPtr;
	Tk_Window tkwin;
 
	if (argc < 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " pathName ?options?\"", (char *) NULL);
		return TCL_ERROR;
	}
	tkwin = Tk_CreateWindowFromPath(interp, main,
			 argv[1], (char *) NULL);
	if (tkwin == NULL) {
		return TCL_ERROR;
	}
	/*
	 * Set resource class.
	 */
	Tk_SetClass(tkwin, "Clock");
	/*
	 * Allocate and initialize the widget record.
	 */
	clockPtr = (Clock *) Tcl_Alloc(sizeof(Clock));
	clockPtr->tkwin = tkwin;
	clockPtr->display = Tk_Display(tkwin);
	clockPtr->interp = interp;
	clockPtr->borderWidth = 0;
	clockPtr->highlightWidth = 0;
	clockPtr->relief = TK_RELIEF_FLAT;
	clockPtr->background = NULL;
	clockPtr->foreground = NULL;
	clockPtr->highlight = NULL;
	clockPtr->highlightBg = NULL;
	clockPtr->tkfont = NULL;
	clockPtr->textGC = None;
	clockPtr->token = NULL;
	clockPtr->clock = NULL;
	clockPtr->format = NULL;
	clockPtr->numChars = 0;
	clockPtr->textWidth = 0;
	clockPtr->textHeight = 0;
	clockPtr->padX = 0;
	clockPtr->padY = 0;
	clockPtr->flags = 0;
	/*
	 * Register a handler for when the window is
	 * exposed or resized.
	 */
	Tk_CreateEventHandler(clockPtr->tkwin,
		ExposureMask|StructureNotifyMask|FocusChangeMask,
		ClockEventProc, (ClientData) clockPtr);
	/*
	 * Create a Tcl command that operates on the widget.
	 */
	clockPtr->widgetCmd = Tcl_CreateCommand(interp,
		Tk_PathName(clockPtr->tkwin),
		ClockInstanceCmd,
		(ClientData) clockPtr, (void (*)()) NULL);
	/*
	 * Parse the command line arguments.
	 */
	if (ClockConfigure(interp, clockPtr,
			 argc-2, argv+2, 0) != TCL_OK) {
		Tk_DestroyWindow(clockPtr->tkwin);
		return TCL_ERROR;
	}
	interp->result = Tk_PathName(clockPtr->tkwin);
	return TCL_OK;
}


